// Book.java      

import swarm.defobj.Zone;
import swarm.objectbase.SwarmObjectImpl;
import java.util.*;
import java.math.BigDecimal;

import swarm.collections.*; 

/**
 * This is the book.
 * The book works on the basis of two arrayList containing sell order in
 * increasing order or buy order in decreasing order. The orders are arrays which 
 * have in the first position the price, in the second the number of agent who places
 * the order and in the third the quantity.
 * If an order obtains an immediate matching, it is not filed 
 * 
 * @author Antonio de Ruvo  e-mail:antderu@libero.it 
 * @author Integrations of Bruno Mencarelli 
 */

public class Book extends SwarmObjectImpl{
  /** The index used for send message to agents. */
  public ListIndex indexAgentListIndex;
  /** If 1 many objects print data on the terminal window. */
  public int printing;
  /** The total number of istant. */
  public int istantNumber;
  /** The number of actual istant. */
  public int numberOfIstant;
  /** The number of orders received from agents. */
  public int orderNumber;
  /** The number of share negotiated */
  public int count=0;
  /** The number of shares in buy side of book. */ 
  public int sharesInBuySide;
  /** The number of shares in sell side of book. */
  public int sharesInSellSide;
  /** The number of shares in buy side of book in opening or closing Auctions. */ 
  public int sharesInBuySideInOpeningOrClosingAuctions;
  /** The number of shares in sell side of book in opening or closing Auctions. */
  public int sharesInSellSideInOpeningOrClosingAuctions;
  /** The price of the last contract closing. */
  public double executedPrice=1;
  /** The mean price of yesterday. */ 
  public double meanPrice=1;
  /** The closing price of yesterday. */
  public double  previousClosingPrice=executedPrice;
  /** The addition of prices*quantity to calculate the mean price. */
  public double currentMeanPrice=0;
  /** The arrayList which contains the buy orders. */
  public ArrayList buyOrderStorehouse=new ArrayList();
  /** The arrayList which contains the sell orders. */
  public ArrayList sellOrderStorehouse=new ArrayList();
  /** The arrayList to log the closing contracts. */
  public ArrayList closingContractLog=new ArrayList();
  /** The arrayList as a table to evaluete the teoric price. */
  public ArrayList tableForTheoreticalPrice=new ArrayList();
  /** The arrayList which contains the buy orders with auction price. */
  public ArrayList buyOrderWithAuctionPriceStorehouse=new ArrayList();
  /** The arrayList which contains the sell orders with auction price. */
  public ArrayList sellOrderWithAuctionPriceStorehouse=new ArrayList();
  /** The reference price of each days. */
  public double referencePrice=1;
  /** The theoretical price. */
  public double theoreticalPrice=0;
  /** The opening price. */
  public double openingPrice=1;
  /** The number of buy order with auction price. */
  public int nOfBuyOrderWithAuctionPrice;
  /** The number of sell order with auction price. */
  public int nOfSellOrderWithAuctionPrice;
  /** The number ofshares with auction price in buy side. */
  public int nOfBuySharesWithAuctionPrice;
  /** The number ofshares with auction price in sell side. */
  public int nOfSellSharesWithAuctionPrice;

  /** The length of the mean price history. */
  public int meanPriceHistoryLength=1000;
  /** The length of the local price history. */
  public int localHistoryLength=1000;
  /** The vector containing the history of mean prices. */
  public double[] meanPriceHistory = new double[meanPriceHistoryLength];
  /** The vector containing the local price history. */
  public double[] localHistory = new double[localHistoryLength];

  /** Constructor for a new book. */
  public Book(Zone aZone){
    super(aZone);
    }

  /** To inform the agent of executed price. 
   * @see BasicSumAgent
   */
  private void message(double n, double p, double q){
    BasicSumAgent anAgent;
    indexAgentListIndex.setOffset((int)n-1);
    anAgent = (BasicSumAgent)indexAgentListIndex.get();
    double[]confirmationOfExecutedPrice={p,q};
    anAgent.setConfirmationOfExecutedPrice(confirmationOfExecutedPrice);         
  }

  /** To standardize the order. */
  private double[] standardizedOrder(double[] orderFromAgent){
    double []  order=new double [5];
    if(orderFromAgent.length==2){
      double[] orderStandard={orderFromAgent[0], orderFromAgent[1], 1, 0, 0};
      order=orderStandard;
    }
    if(orderFromAgent.length==3){
      double [] orderStandard={orderFromAgent[0], orderFromAgent[1],
                               orderFromAgent[2], 0, 0};
      order=orderStandard;
    }
    if(orderFromAgent.length==4){
      double [] orderStandard={orderFromAgent[0], orderFromAgent[1],
                               orderFromAgent[2], orderFromAgent[3], 0};
      order=orderStandard;
    }
    return(order);      
  }

  /** To extract data from the orders received from agents. */
  private double extract( ArrayList nameList, int indexOrder, int positionDatum ){
    Object[] array=nameList.toArray();
    double datum=((double []) array[indexOrder])[positionDatum];
    return(datum);
  }

  /** To fill sellOrderStorehouse in increasing order */
  private int fillStorehouseIncreasingP(double newPrice){
    int index=0;
    for(int i=0; i<sellOrderStorehouse.size(); i++){
      if(newPrice<extract(sellOrderStorehouse, i, 0)){
	index=i;
	break;
      }
    }	
    return(index);    
  }

  /** To fill buyOrderStorehouse in descreasing order. */
  private int fillStorehouseDecreasingP (double newPrice){
    int index=0;
    for(int i=0; i<buyOrderStorehouse.size(); i++){
      if(newPrice>extract(buyOrderStorehouse, i, 0)){
	index=i;
	break;
      }
    }
    return(index);
  }

  /** This is the SetLocal method, used to shift the rows of the localHistory vector.
   * @author Mencarelli
   */ 
  public void setLocal( double[] vect,double p ){
  int i;
  if(p==0) return;
  for(i=vect.length-2;i>=0;i--){
      vect[i+1]=vect[i];}
  vect[0]=p;
  }

  /** This is the getLocal method, that counts for the difference
   *  between the buying and selling orders sent by the agents.
   * @author Mencarelli
   */
  public int getLocal( double[] vect ){
    int i,tot=0;
    for(i=0;i<vect.length;i++){
      if(vect[i]>0) tot++;
      if(vect[i]<0) tot--;
      }
  return tot;
  }

  /** To set the value of option printing. */
  public void setPrinting(int p){
    printing=p;
  }

  /** To set the value of total number of istant */
  public void setIstantNumber(int n){
    istantNumber=n;
  }

  /** To know the adress of memory of the agent. */
  public void setAgentListIndex(ListIndex i){
    indexAgentListIndex=i;
    }

  /** To set the value of actual istant */
  public void setNumberOfIstant(){
    numberOfIstant++;
  }

  /** At the end of each day to calculate the mean price. */
  public void setMeanPrice(){
    if(count>0)
      meanPrice=currentMeanPrice/count; 
    // otherwise we keep previous value
  }
  
  /** At the end of each day to calculate the reference price. */
  public void setRefencePrice(){
    if(previousClosingPrice!=0)
      referencePrice=previousClosingPrice;
    if(previousClosingPrice==0){
      // The price's weighed average of last 10% quantity 
      double quantity=count*0.1;
      double q=0;
      double weighedAverage=0;
      Object[] matrix=closingContractLog.toArray();
      int index=matrix.length;
      while(q<quantity){
	index--; 
	weighedAverage+=(((double[])matrix[index])[0]*((double[])matrix[index])[1]);
        q+=((double[])matrix[index])[0];
      }
      weighedAverage-=(q-quantity)*((double[])matrix[index])[1];
      referencePrice=weighedAverage/quantity;
      if(printing==1){
        System.out.println("The log of closing contrancts is:");
        for(int i=0; i<closingContractLog.size(); i++){
          for(int y=0; y<2; y++){
            System.out.print(extract(closingContractLog, i, y) + " ");
          }      
          System.out.println();
        };
        System.out.println();
	System.out.println("count = " + count );
	System.out.println("The quantity of last 10% is " + quantity );
	System.out.println("weighedAverage = " + weighedAverage );
	System.out.println("ReferencePrice = " + referencePrice);
      }
    }
  }

  /** At the beginning of each day to clean the book. */
  public void setClean(){
    sellOrderStorehouse.clear();
    buyOrderStorehouse.clear();
    closingContractLog.clear();
    tableForTheoreticalPrice.clear();
    buyOrderWithAuctionPriceStorehouse.clear();
    sellOrderWithAuctionPriceStorehouse.clear();
    previousClosingPrice=openingPrice; // the closing price of "yesterday"
    orderNumber=0;
    for(int i=meanPriceHistoryLength-2;i>=0;i--){
      meanPriceHistory[i+1]=meanPriceHistory[i];
    }
    meanPriceHistory[0]=meanPrice;
    currentMeanPrice=0;
    count=0;
    sharesInSellSide=0;
    sharesInBuySide=0;
    nOfBuyOrderWithAuctionPrice=0;
    nOfSellOrderWithAuctionPrice=0;
    nOfBuySharesWithAuctionPrice=0;
    nOfSellSharesWithAuctionPrice=0;
    numberOfIstant=0;
    if(printing==1)
	System.out.println("Book: clean");
  }

 /** To set rows in tableForTheoreticalPrice. */
  private double[] setRows(double[] order){
    double[] row=new double[5];
    if(order[3]==0){
      if(order[0]>0){
        row[0]=order[0];
        row[1]=order[2];
	row[3]=Math.min(row[1],row[2]);
	row[4]=Math.abs(row[1]-row[2]);
      }
      else if(order[0]<0){
        row[0]=-order[0];
        row[2]=order[2];
	row[3]=Math.min(row[1],row[2]);
	row[4]=Math.abs(row[1]-row[2]);
      }
    }
    if(order[3]==1){
      row[0]=0;
      row[1]=order[2];
      row[3]=Math.min(row[1],row[2]);
      row[4]=Math.abs(row[1]-row[2]);
    }
    else if(order[3]==2){
      row[0]=0;
      row[2]=order[2];
      row[3]=Math.min(row[1],row[2]);
      row[4]=Math.abs(row[1]-row[2]);
    }
    return(row);
  }

  /** To set the prices in decreasing order for evaluete the theoretical price. */
  private void setTableForTheoreticalPrice(double[] order){
    double[] row;
    row=setRows(order);
    Object[] matrix=tableForTheoreticalPrice.toArray();

    if(tableForTheoreticalPrice.size()==0){
      tableForTheoreticalPrice.add(row);
      if(row[1]>0 && row[0]==0)
        nOfBuySharesWithAuctionPrice+=row[1];
      if(row[2]>0 && row[0]==0)
        nOfSellSharesWithAuctionPrice+=row[2];
    }

    else if(row[0]==0){
      if(row[1]>0)
        nOfBuySharesWithAuctionPrice+=row[1];
      if(row[2]>0)
        nOfSellSharesWithAuctionPrice+=row[2];
      for(int i=0; i<matrix.length; i++){
        ((double[])matrix[i])[1]+=row[1];
        ((double[])matrix[i])[2]+=row[2];
        ((double[])matrix[i])[3]=Math.min(((double[])matrix[i])[1], ((double[])matrix[i])[2]);
        ((double[])matrix[i])[4]=Math.abs(((double[])matrix[i])[1]-((double[])matrix[i])[2]);
        tableForTheoreticalPrice.set(i, matrix[i]);
      }
    }

    else if(row[0]>0){
      boolean thereIs=false;
      for(int i=0; i<matrix.length; i++){
        if(row[0]==((double[])matrix[i])[0]){
          thereIs=true;
        }
      }
      if(thereIs==false){
        if(row[0]>((double[])matrix[0])[0] && ((double[])matrix[0])[0]!=0){
	  double[] newRow=new double[5];
          newRow[0]=row[0];
          newRow[1]=nOfBuySharesWithAuctionPrice;
	  newRow[2]=((double[])matrix[0])[2];
          tableForTheoreticalPrice.add(0, newRow);
	}
        if(row[0]>((double[])matrix[0])[0] && ((double[])matrix[0])[0]==0){
	  ((double[])matrix[0])[0]=row[0];
          tableForTheoreticalPrice.set(0, matrix[0]);
	}
	else if(row[0]<((double[])matrix[0])[0]){
          if(row[0]<((double[])matrix[matrix.length-1])[0]){
	    double[] newRow=new double[5];
            newRow[0]=row[0];
	    newRow[1]=((double[])matrix[0])[1];
	    newRow[2]=nOfSellSharesWithAuctionPrice;
            tableForTheoreticalPrice.add(newRow);
	  }
	  else{
            for(int i=0; i<matrix.length; i++){
              if(row[0]>((double[])matrix[i])[0]){
	        double[] newRow=new double[5];
                newRow[0]=row[0];
                newRow[1]=((double[])matrix[i-1])[1];
                newRow[2]=((double[])matrix[i])[2];
                tableForTheoreticalPrice.add(i, newRow);
	        break;
	      }
            }
	  }
	}
      }

      matrix=tableForTheoreticalPrice.toArray();
      if(row[1]>0){
        for(int i=0; i<matrix.length; i++){
	  if(row[0]>=((double[])matrix[i])[0]){
            ((double[])matrix[i])[1]+=row[1];
            ((double[])matrix[i])[2]+=row[2];
            ((double[])matrix[i])[3]=Math.min(((double[])matrix[i])[1],
                                              ((double[])matrix[i])[2]);
            ((double[])matrix[i])[4]=Math.abs(((double[])matrix[i])[1]-
                                              ((double[])matrix[i])[2]);
            tableForTheoreticalPrice.set(i, matrix[i]);
	  }
	}
      }

      else if(row[2]>0){
        for(int i=0; i<matrix.length; i++){
	  if(row[0]<=((double[])matrix[i])[0]){
            ((double[])matrix[i])[1]+=row[1];
            ((double[])matrix[i])[2]+=row[2];
            ((double[])matrix[i])[3]=Math.min(((double[])matrix[i])[1],
                                              ((double[])matrix[i])[2]);
            ((double[])matrix[i])[4]=Math.abs(((double[])matrix[i])[1]-
                                              ((double[])matrix[i])[2]);
            tableForTheoreticalPrice.set(i, matrix[i]);
	  }
	}
      }
    }
  }

  /** To set the theoretical price. */
  public void setTheoreticalPrice(){
    int numberOfEqualMaxNegotiableQuantity=0;
    int numberOfEqualMinDifferenceBetweenVolume=0;
    double maxQuantityNegotiable=0;
    double minDifferenceBetweenVolume=0;
    int index=0;
    int index1=0;
    int index2=0;
    if(printing==1){
      System.out.println();
      System.out.println("The table for theoretical price is:");
      for(int i=0; i<tableForTheoreticalPrice.size(); i++){
        for(int y=0; y<5; y++){
          System.out.print(extract(tableForTheoreticalPrice, i, y) + " ");
        }      
        System.out.println();
      };
      System.out.println();
      System.out.println("The reference price is: " + referencePrice);
      System.out.println();
    }
    Object[] matrix=tableForTheoreticalPrice.toArray();
    //first rule
    for(int i=0;i<matrix.length;i++){
      if(((double[])matrix[i])[3]>maxQuantityNegotiable){
	numberOfEqualMaxNegotiableQuantity=1;
        maxQuantityNegotiable=((double[])matrix[i])[3];
	index=i;  
      }
      else if(((double[])matrix[i])[3]==maxQuantityNegotiable)
	numberOfEqualMaxNegotiableQuantity++;
    }
    if(matrix.length==0 || maxQuantityNegotiable==0 )
      theoreticalPrice=0;
    else if(numberOfEqualMaxNegotiableQuantity==1)
      theoreticalPrice=((double[])matrix[index])[0];
    //second rule
    else if(numberOfEqualMaxNegotiableQuantity>1){
      index1=index;
      minDifferenceBetweenVolume=((double[])matrix[index])[4];
      numberOfEqualMinDifferenceBetweenVolume=1;
      for(int i=index+1;i<index+numberOfEqualMaxNegotiableQuantity;i++){
        if(((double[])matrix[i])[4]<minDifferenceBetweenVolume){
          numberOfEqualMinDifferenceBetweenVolume=1;
          minDifferenceBetweenVolume=((double[])matrix[i])[4];
   	  index1=i;  
        }
        else if(((double[])matrix[i])[4]==minDifferenceBetweenVolume)
	  numberOfEqualMinDifferenceBetweenVolume++;
      }
      if(numberOfEqualMinDifferenceBetweenVolume==1)
        theoreticalPrice=((double[])matrix[index1])[0];
      //third rule
      if(numberOfEqualMinDifferenceBetweenVolume>1){
        double[] array=new double[numberOfEqualMinDifferenceBetweenVolume];
        for(int i=0;i<array.length;i++){
	  array[i]=Math.abs(((double[])matrix[i+index1])[0]-referencePrice);
	}
	double min=array[0];
	int numberOfEqualMin=1;
	index2=index1;
	for(int i=0;i<array.length-1;i++){
	  if(array[i]>array[i+1]){
	    numberOfEqualMin=1;
	    min=array[i+1];
	    index2=index1+i+1;
	  }
	  if(array[i]<array[i+1]){
	    numberOfEqualMin=1;
	    min=array[i];
	    index2=index1+i;
	  }
	  else if(array[i]==array[i+1])
	    numberOfEqualMin=2;
	}
	if(numberOfEqualMin==1){
	  if(((double[])matrix[index2])[0]<referencePrice)
	    theoreticalPrice=referencePrice-min;
	  else
	    theoreticalPrice=referencePrice+min;
	}
	//fourth rule
	else if(numberOfEqualMin==2)
	  theoreticalPrice=referencePrice+min;
      }
    }
    if(buyOrderStorehouse.size()==0 && buyOrderWithAuctionPriceStorehouse.size()!=0 &&
       sellOrderStorehouse.size()==0 && sellOrderWithAuctionPriceStorehouse.size()!=0){
       theoreticalPrice=referencePrice;
    }
    if(printing==1){
      System.out.println("The theoretical price is: " + theoreticalPrice);
      System.out.println();
    }
  }

  /** To set the opening price. */
  public void setOpeningPrice(){
    if(theoreticalPrice==0)
      openingPrice=0;
    else if(theoreticalPrice>0){
      //validation
      if(theoreticalPrice<=referencePrice*1.1 && theoreticalPrice>=referencePrice*0.9)
        openingPrice=theoreticalPrice;
      else
	openingPrice=0;
    }
    if(printing==1){
      if(numberOfIstant==0)
        System.out.println("The opening price is: " + openingPrice);
      else if(numberOfIstant==istantNumber)
        System.out.println("The closing price is: " + openingPrice);
      System.out.println();
      System.out.println("Buy orders with auction Price: ");
      for(int i =0; i<buyOrderWithAuctionPriceStorehouse.size(); i++){
        for(int y=0; y<5; y++){
          System.out.print(extract(buyOrderWithAuctionPriceStorehouse, i, y) + " ");
        }      
        System.out.println();
      };
      System.out.println();
      System.out.println("Sell orders with auction Price: ");
      for(int i =0; i<sellOrderWithAuctionPriceStorehouse.size(); i++){
        for(int y=0; y<5; y++){
          System.out.print(extract(sellOrderWithAuctionPriceStorehouse, i, y) + " ");
        }      
        System.out.println();
      };
      System.out.println();
    }
  }

  /** To conclude contracts (if opening price > 0) or to fill the order with auction price. */
  public void opening(){
    Object[] matrix=buyOrderWithAuctionPriceStorehouse.toArray();
    Object[] matrix1=sellOrderWithAuctionPriceStorehouse.toArray();
    if(openingPrice!=0){
      referencePrice=openingPrice;
      for(int i=0;i<matrix.length;i++){
	((double[])matrix[i])[0]=openingPrice;
	((double[])matrix[i])[3]=0;
      }
      for(int i=0;i<matrix1.length;i++){
	((double[])matrix1[i])[0]=openingPrice;
	((double[])matrix1[i])[3]=0;
      }
    }
    //If the auction falied, there is not an other auction
    else if(openingPrice==0 && theoreticalPrice>0 &&
           (theoreticalPrice>referencePrice*1.1 || theoreticalPrice<referencePrice*0.9)){
      if(buyOrderStorehouse.size()!=0 && sellOrderStorehouse.size()!=0){
        for(int i=0;i<matrix.length;i++){
  	  ((double[])matrix[i])[0]=extract(buyOrderStorehouse,0,0);
	  ((double[])matrix[i])[3]=0;
        }
        for(int i=0;i<matrix1.length;i++){
	  ((double[])matrix1[i])[0]=extract(sellOrderStorehouse,0,0);
	  ((double[])matrix1[i])[3]=0;
        }
      }
      else if(buyOrderStorehouse.size()!=0 && sellOrderStorehouse.size()==0){
        for(int i=0;i<matrix.length;i++){
  	  ((double[])matrix[i])[0]=extract(buyOrderStorehouse,0,0);
	  ((double[])matrix[i])[3]=0;
        }
        for(int i=0;i<matrix1.length;i++){
	  ((double[])matrix1[i])[0]=referencePrice;
	  ((double[])matrix1[i])[3]=0;
        }
      }
      if(buyOrderStorehouse.size()==0 && sellOrderStorehouse.size()!=0){
        for(int i=0;i<matrix.length;i++){
  	  ((double[])matrix[i])[0]=referencePrice;
	  ((double[])matrix[i])[3]=0;
        }
        for(int i=0;i<matrix1.length;i++){
	  ((double[])matrix1[i])[0]=extract(sellOrderStorehouse,0,0);
	  ((double[])matrix1[i])[3]=0;
        }
      }
    }
    //If one side is empty or there are only orders with auction price
    else if(openingPrice==0 && buyOrderStorehouse.size()==0 && matrix.length!=0 &&
                               sellOrderStorehouse.size()==0 && matrix1.length==0){
      for(int i=0;i<matrix.length;i++){
	((double[])matrix[i])[0]=referencePrice;
	((double[])matrix[i])[3]=0;
      }
    }
    else if(openingPrice==0 && buyOrderStorehouse.size()==0 && matrix.length==0 &&
                               sellOrderStorehouse.size()==0 && matrix1.length!=0){
      for(int i=0;i<matrix1.length;i++){
	((double[])matrix1[i])[0]=referencePrice;
	((double[])matrix1[i])[3]=0;
      }
    }
    else if(openingPrice==0 && buyOrderStorehouse.size()!=0 && matrix.length!=0 &&
                               sellOrderStorehouse.size()==0 && matrix1.length==0){
      for(int i=0;i<matrix.length;i++){
	((double[])matrix[i])[0]=extract(buyOrderStorehouse,0,0);
	((double[])matrix[i])[3]=0;
      }
    }
    else if(openingPrice==0 && buyOrderStorehouse.size()==0 && matrix.length==0 &&
                               sellOrderStorehouse.size()!=0 && matrix1.length!=0){
      for(int i=0;i<matrix1.length;i++){
	((double[])matrix1[i])[0]=extract(sellOrderStorehouse,0,0);
	((double[])matrix1[i])[3]=0;
      }
    }
    if(printing==1){
      System.out.println("Correct buy orders with auction Price: ");
      for(int i =0; i<buyOrderWithAuctionPriceStorehouse.size(); i++){
        for(int y=0; y<5; y++){
          System.out.print(extract(buyOrderWithAuctionPriceStorehouse, i, y) + " ");
        }      
        System.out.println();
      };
      System.out.println();
      System.out.println("Correct sell orders with auction Price: ");
      for(int i =0; i<sellOrderWithAuctionPriceStorehouse.size(); i++){
        for(int y=0; y<5; y++){
          System.out.print(extract(sellOrderWithAuctionPriceStorehouse, i, y) + " ");
        }      
        System.out.println();
      };
      System.out.println();
    }
    //To fill the orders in buy side
    double price=0;
    if(matrix.length!=0)
      price=((double[])matrix[0])[0];
    Object[] matrix2=buyOrderStorehouse.toArray();
    int index=0;
    for(int i=0; i<matrix2.length; i++){
      if(price>=((double[])matrix2[i])[0]){
	index=i;
	break;
      }
    }
    if(matrix2.length==0){
      for(int i=0;i<matrix.length;i++){
        buyOrderStorehouse.add((double[])matrix[i]);
      }
    }
    else if(matrix2.length!=0){
      if(index==0 && price<((double[])matrix2[matrix2.length-1])[0]){
        for(int i=0;i<matrix.length;i++){
          buyOrderStorehouse.add((double[])matrix[i]);
        }
      }
      else if(((double[])matrix2[index])[0]==price){
        for(int i=0;i<matrix.length;i++){
          if(((double[])matrix[i])[4]<extract(buyOrderStorehouse,index,4)){
            buyOrderStorehouse.add(index,(double[])matrix[i]);
	    index++;
          }
  	  else if(((double[])matrix[i])[4]>extract(buyOrderStorehouse,index,4)){
	    buyOrderStorehouse.add(index+1,(double[])matrix[i]);
	    index++;
          }
        }
      }
      else if(matrix2.length!=0 && ((double[])matrix2[index])[0]>price){
        for(int i=0;i<matrix.length;i++){
           buyOrderStorehouse.add(index,(double[])matrix[i]);
	   index++;
        }
      }
      else if(matrix2.length!=0 && ((double[])matrix2[index])[0]<price){
        for(int i=0;i<matrix.length;i++){
           buyOrderStorehouse.add(index,(double[])matrix[i]);
	   index++;
        }
      }
    }
    if(printing==1) {
      System.out.println();
      System.out.println("All buy side: ");
      for(int i=0; i<buyOrderStorehouse.size(); i++){
        for(int y=0; y<5; y++){
          System.out.print(extract(buyOrderStorehouse, i, y) + " ");
        }      
        System.out.println();
      };
      System.out.println();
    }
    //To fill the orders in sell side
    if(matrix1.length!=0)
      price=((double[])matrix1[0])[0];
    Object[] matrix3=sellOrderStorehouse.toArray();
    int index1=0;
    for(int i=0; i<matrix3.length; i++){
      if(price<=((double[])matrix3[i])[0]){
	index1=i;
	break;
      }
    }
    if(matrix3.length==0){
      for(int i=0;i<matrix1.length;i++){
        sellOrderStorehouse.add((double[])matrix1[i]);
      }
    }
    else if(matrix3.length!=0){
      if(index1==0 && price>((double[])matrix3[matrix3.length-1])[0]){
        for(int i=0;i<matrix1.length;i++){
          sellOrderStorehouse.add((double[])matrix1[i]);
        }
      }
      else if(((double[])matrix3[index1])[0]==price){
        for(int i=0;i<matrix1.length;i++){
          if(((double[])matrix1[i])[4]<extract(sellOrderStorehouse,index1,4)){
            sellOrderStorehouse.add(index1,(double[])matrix1[i]);
	    index1++;
          }
  	  else if(((double[])matrix1[i])[4]>extract(sellOrderStorehouse,index1,4)){
	    sellOrderStorehouse.add(index1+1,(double[])matrix1[i]);
	    index1++;
          }
        }
      }
      else if(matrix3.length!=0 && ((double[])matrix3[index1])[0]<price){
        for(int i=0;i<matrix1.length;i++){
           sellOrderStorehouse.add(index1,(double[])matrix1[i]);
	   index1++;
        }
      }
      else if(matrix3.length!=0 && ((double[])matrix3[index1])[0]>price){
        for(int i=0;i<matrix1.length;i++){
           sellOrderStorehouse.add(index1,(double[])matrix1[i]);
	   index1++;
        }
      }
    }
    if(printing==1) {
      System.out.println();
      System.out.println("All sell side: ");
      for(int i=0; i<sellOrderStorehouse.size(); i++){
        for(int y=0; y<5; y++){
          System.out.print(extract(sellOrderStorehouse, i, y) + " ");
        }      
        System.out.println();
      };
      System.out.println();
    }

    //To close the contracts
    if(openingPrice!=0){
      while(buyOrderStorehouse.size()!=0 && sellOrderStorehouse.size()!=0 &&
            extract(buyOrderStorehouse, 0, 0)>=openingPrice &&
            extract(sellOrderStorehouse, 0, 0)<=openingPrice){
        if(extract(sellOrderStorehouse, 0, 2)==extract(buyOrderStorehouse, 0, 2)){ 
          count+=extract(buyOrderStorehouse, 0, 2);
          executedPrice=openingPrice;
          message(extract(sellOrderStorehouse, 0, 1), -executedPrice,
                  extract(sellOrderStorehouse, 0, 2));
          message(extract(buyOrderStorehouse, 0, 1), executedPrice,
                  extract(buyOrderStorehouse, 0, 2));
          double[] log=new double[2];
          log[0]=extract(buyOrderStorehouse, 0, 2);
          log[1]=executedPrice;
          closingContractLog.add(log);
          currentMeanPrice+=executedPrice*extract(buyOrderStorehouse, 0, 2);
          sharesInBuySide-=extract(buyOrderStorehouse, 0, 2);
          sharesInSellSide-=extract(sellOrderStorehouse, 0, 2);
          buyOrderStorehouse.remove(0);
          sellOrderStorehouse.remove(0);
          if(printing==1) {
            System.out.println("Buy order storehouse:");
            for(int i=0; i<buyOrderStorehouse.size(); i++){
              for(int y=0; y<5; y++){
                System.out.print(extract(buyOrderStorehouse, i, y) + " ");
              }      
              System.out.println();
            };
            System.out.println();
            System.out.println("Sell order storehouse:");
            for(int i=0; i<sellOrderStorehouse.size(); i++){
              for(int y=0; y<5; y++){
                System.out.print(extract(sellOrderStorehouse, i, y) + " ");
              }      
              System.out.println();
            };
            System.out.println();
          }
        }
        else if(extract(sellOrderStorehouse, 0, 2)<extract(buyOrderStorehouse, 0, 2)){
          count+=extract(sellOrderStorehouse, 0, 2);
          executedPrice=openingPrice;
          message(extract(sellOrderStorehouse, 0, 1), -executedPrice,
                  extract(sellOrderStorehouse, 0, 2));
          message(extract(buyOrderStorehouse, 0, 1), executedPrice, 
                  extract(sellOrderStorehouse, 0, 2));
          double[] log=new double[2];
          log[0]=extract(sellOrderStorehouse, 0, 2);
          log[1]=executedPrice;
          closingContractLog.add(log);
          currentMeanPrice+=executedPrice*extract(sellOrderStorehouse, 0, 2);
          ((double[]) buyOrderStorehouse.get(0))[2]=extract(buyOrderStorehouse, 0, 2)
                                                    -extract(sellOrderStorehouse, 0, 2);
          sharesInBuySide-=extract(sellOrderStorehouse, 0, 2);
          sharesInSellSide-=extract(sellOrderStorehouse, 0, 2);
          sellOrderStorehouse.remove(0);
          if(printing==1) {
            System.out.println("Buy order storehouse:");
            for(int i=0; i<buyOrderStorehouse.size(); i++){
              for(int y=0; y<5; y++){
                System.out.print(extract(buyOrderStorehouse, i, y) + " ");
              }      
              System.out.println();
            };
            System.out.println();
            System.out.println("Sell order storehouse:");
            for(int i=0; i<sellOrderStorehouse.size(); i++){
              for(int y=0; y<5; y++){
                System.out.print(extract(sellOrderStorehouse, i, y) + " ");
              }      
              System.out.println();
            };
            System.out.println();
          }
        }
	else if(extract(sellOrderStorehouse, 0, 2)>extract(buyOrderStorehouse, 0, 2)){
          count+=extract(buyOrderStorehouse, 0, 2);
          executedPrice=openingPrice;
          message(extract(sellOrderStorehouse, 0, 1), -executedPrice,
                  extract(buyOrderStorehouse, 0, 2));
          message(extract(buyOrderStorehouse, 0, 1), executedPrice,
                  extract(buyOrderStorehouse, 0, 2));
          double[] log=new double[2];
          log[0]=extract(buyOrderStorehouse, 0, 2);
          log[1]=executedPrice;
          closingContractLog.add(log);
          currentMeanPrice+=executedPrice*extract(buyOrderStorehouse, 0, 2);
          ((double[]) sellOrderStorehouse.get(0))[2]=
                  ((double[]) sellOrderStorehouse.get(0))[2]-extract(buyOrderStorehouse, 0, 2);
          sharesInBuySide-=extract(buyOrderStorehouse, 0, 2);
          sharesInSellSide-=extract(buyOrderStorehouse, 0, 2);
          buyOrderStorehouse.remove(0);
          if(printing==1) {
            System.out.println("Buy order storehouse:");
            for(int i=0; i<buyOrderStorehouse.size(); i++){
              for(int y=0; y<5; y++){
                System.out.print(extract(buyOrderStorehouse, i, y) + " ");
              }      
              System.out.println();
            };
            System.out.println();
            System.out.println("Sell order storehouse:");
            for(int i=0; i<sellOrderStorehouse.size(); i++){
              for(int y=0; y<5; y++){
                System.out.print(extract(sellOrderStorehouse, i, y) + " ");
              }      
              System.out.println();
            };
            System.out.println();
          }
	}
      }
    }
  }

  //** To recreate the table for theoretical price. */
  public void recreateTableForTheoreticalPrice(){
    tableForTheoreticalPrice.clear();
    buyOrderWithAuctionPriceStorehouse.clear();
    sellOrderWithAuctionPriceStorehouse.clear();
    nOfBuyOrderWithAuctionPrice=0;
    nOfSellOrderWithAuctionPrice=0;
    nOfBuySharesWithAuctionPrice=0;
    nOfSellSharesWithAuctionPrice=0;
    Object[] matrix=sellOrderStorehouse.toArray();
    Object[] matrix1=buyOrderStorehouse.toArray();
    for(int i=0; i<matrix.length; i++){
      ((double[])matrix[i])[0]=-((double[])matrix[i])[0];
      setTableForTheoreticalPrice(((double[])matrix[i]));
      ((double[])matrix[i])[0]=-((double[])matrix[i])[0];
    }
    for(int i=0; i<matrix1.length; i++){
      setTableForTheoreticalPrice(((double[])matrix1[i]));
    }
    if(printing==1){
      System.out.println();
      System.out.println("The table for theoretical price is:");
      for(int i=0; i<tableForTheoreticalPrice.size(); i++){
        for(int y=0; y<5; y++){
          System.out.print(extract(tableForTheoreticalPrice, i, y) + " ");
        }      
        System.out.println();
      };
      System.out.println();
    }
  }

  /** Receiving an order in opening or closing Auctions from an agent. */
  public void setOrderInOpeningOrClosingAuctionsFromAgent(double[] orderFromAgent) {
    double [] order = standardizedOrder(orderFromAgent);
    orderNumber++;
    order[4]=orderNumber;
    order[0] = new BigDecimal(order[0]).setScale(3, BigDecimal.ROUND_HALF_DOWN).doubleValue();

    /* local History. */
    setLocal(localHistory,order[0]);

    if(printing==1){
      if(numberOfIstant==0){
        System.out.println("The book received a before opening order from agent " + order[1] +
        " at price " + order[0] + " for " + order[2]+ " share(s), with option = " + order[3] +
        " number's order " + order[4]);
      }
      else if(numberOfIstant==istantNumber){
        System.out.println("The book received a before closing order from agent " + order[1] +
        " at price " + order[0] + " for " + order[2]+ " share(s), with option = " + order[3] +
        " number's order " + order[4]);
      }
    }

  /* if  price=0 and it is not an order with auction price or quantity==0 no action required,
   * but sending a  message to the agent with price=0 */
    if(order[0]==0 && order[3]==0 || order[2]==0)
       message(order[1], 0.0, 0.0);

  // the agent is selling at minimum price if  price<0
     else if(order[0]<0 && order[3]==0) {
       message(order[1], 0.0, 0.0);
       setTableForTheoreticalPrice(order);
       order[0]=-order[0];
  /* to insert the order in the book in increasing order. 
   * With add( )the arrayList updates the index */
       if(sellOrderStorehouse.size()==0 ||
          order[0]>=extract(sellOrderStorehouse, (sellOrderStorehouse.size()-1), 0)) 
         sellOrderStorehouse.add(order);
       else
	 sellOrderStorehouse.add(fillStorehouseIncreasingP(order[0]), order);
       sharesInSellSide+=order[2];
       sharesInSellSideInOpeningOrClosingAuctions=sharesInSellSide;
       if(printing==1) {
       System.out.println("Sell order storehouse:");
         for(int i =0; i<sellOrderStorehouse.size(); i++){
           for(int y=0; y<order.length; y++){
             System.out.print(extract(sellOrderStorehouse, i, y) + " ");
           }      
           System.out.println();
         };
         System.out.println();
       }
     }

  // the agent is buying at the maximun price if price>0
     else if(order[0]>0 && order[3]==0) {
       message(order[1], 0.0, 0.0);
       setTableForTheoreticalPrice(order);
       /* to insert the order in the book in decreasing order. 
        * With add()the arrayList updates the index */
       if(buyOrderStorehouse.size()==0 ||
          order[0]<=extract(buyOrderStorehouse, (buyOrderStorehouse.size()-1) ,0)) 
         buyOrderStorehouse.add(order);
       else
	 buyOrderStorehouse.add(fillStorehouseDecreasingP(order[0]),order);
       sharesInBuySide+=order[2]; 
       sharesInBuySideInOpeningOrClosingAuctions=sharesInBuySide;    
       if(printing==1) {
       System.out.println("Buy order storehouse:");
         for(int i=0; i<buyOrderStorehouse.size(); i++){
           for(int y=0; y<order.length; y++){
             System.out.print(extract(buyOrderStorehouse, i, y) + " ");
           }      
           System.out.println();
         };
         System.out.println();
       }
     }

  // the agent is buying at the auction price
    else if(order[3]==1 && order[0]==0){
      nOfBuyOrderWithAuctionPrice++;
      message(order[1], 0.0, 0.0);
      setTableForTheoreticalPrice(order);
      buyOrderWithAuctionPriceStorehouse.add(order);
      sharesInBuySide+=order[2]; 
      sharesInBuySideInOpeningOrClosingAuctions=sharesInBuySide;
    }

  // the agent is selling at the auction price
    else if(order[3]==2 && order[0]==0){
      nOfSellOrderWithAuctionPrice++;
      message(order[1], 0.0, 0.0);
      setTableForTheoreticalPrice(order);
      sellOrderWithAuctionPriceStorehouse.add(order);
      sharesInSellSide+=order[2];
      sharesInSellSideInOpeningOrClosingAuctions=sharesInSellSide;
    }
  }

  /** Receiving an order when the market is open. */
  public void setOrderFromAgent(double[] orderFromAgent) {
    double [] order = standardizedOrder(orderFromAgent);
    orderNumber++;
    order[4]=orderNumber;
    order[0] = new BigDecimal(order[0]).setScale(3, BigDecimal.ROUND_HALF_DOWN).doubleValue();

    /* local History. */
    if(order[0]!=0)
      setLocal(localHistory,order[0]);

    if(printing==1)
      System.out.println("The book received an order from agent " + order[1] +
      " at price " + order[0] + " for " + order[2]+ " share(s), with option = " + order[3]+
      " number's order " + order[4]);

  /* if  price==0 or quantity==0 no action required, but sending a  message to the agent 
   * with price=0 */
    if(order[0]==0 && order[3]==0 || order[2]==0)
      message(order[1], 0.0, 0.0);

  // the agent is selling at minimum price if  price<0
     else if(order[0]<0 && order[3]==0) {
       /* if there is a compatible order, this one is removed and the order received  
        * it's not inserted in the book. It's not necessary to shift rows because
        * with remove() the arrayList updates the index */
       while(buyOrderStorehouse.size()>0 && extract(buyOrderStorehouse, 0, 0)>=-order[0]
             && order[2]!=0) {
	 if(order[2]==extract(buyOrderStorehouse, 0, 2)){ 
           count+=order[2];
           executedPrice=extract(buyOrderStorehouse, 0, 0);
	   message(order[1], -executedPrice, order[2]);
	   message(extract(buyOrderStorehouse, 0, 1), executedPrice, order[2]);
	   double[] log=new double[2];
	   log[0]=order[2];
	   log[1]=executedPrice;
	   closingContractLog.add(log);
	   currentMeanPrice+=executedPrice*order[2];
           buyOrderStorehouse.remove(0);
           sharesInBuySide-=order[2];
	   order[2]=0;
           if(printing==1) {
             System.out.println("Buy order storehouse:");
             for(int i=0; i<buyOrderStorehouse.size(); i++){
               for(int y=0; y<order.length; y++){
                 System.out.print(extract(buyOrderStorehouse, i, y) + " ");
               }      
               System.out.println();
             };
             System.out.println();
           }
	 }
	 else if(order[2]<extract(buyOrderStorehouse, 0, 2)){
           count+=order[2];
           executedPrice=extract(buyOrderStorehouse, 0, 0);
	   message(order[1], -executedPrice, order[2]);
	   message(extract(buyOrderStorehouse, 0, 1), executedPrice, order[2]);
	   double[] log=new double[2];
	   log[0]=order[2];
	   log[1]=executedPrice;
	   closingContractLog.add(log);
	   currentMeanPrice+=executedPrice*order[2];
	   ((double[]) buyOrderStorehouse.get(0))[2]=extract(buyOrderStorehouse, 0, 2)
                                                     -order[2];
           sharesInBuySide-=order[2];
	   order[2]=0;
           if(printing==1) {
             System.out.println("Buy order storehouse:");
             for(int i=0; i<buyOrderStorehouse.size(); i++){
               for(int y=0; y<order.length; y++){
                 System.out.print(extract(buyOrderStorehouse, i, y) + " ");
               }      
               System.out.println();
             };
             System.out.println();
           }
	 }
	 else if(order[2]>extract(buyOrderStorehouse, 0, 2)){
           count+=extract(buyOrderStorehouse, 0, 2);
           executedPrice=extract(buyOrderStorehouse, 0, 0);
	   message(order[1], -executedPrice, extract(buyOrderStorehouse, 0, 2));
	   message(extract(buyOrderStorehouse, 0, 1), executedPrice,
                     extract(buyOrderStorehouse, 0, 2));
	   double[] log=new double[2];
	   log[0]=extract(buyOrderStorehouse, 0, 2);
	   log[1]=executedPrice;
	   closingContractLog.add(log);
	   currentMeanPrice+=executedPrice*extract(buyOrderStorehouse, 0, 2);
	   order[2]=order[2]-extract(buyOrderStorehouse, 0, 2);
           sharesInBuySide-=extract(buyOrderStorehouse, 0, 2);
           buyOrderStorehouse.remove(0);
           if(printing==1) {
             System.out.println("Buy order storehouse:");
             for(int i=0; i<buyOrderStorehouse.size(); i++){
               for(int y=0; y<order.length; y++){
                 System.out.print(extract(buyOrderStorehouse, i, y) + " ");
               }      
               System.out.println();
             };
             System.out.println();
           }
	 }	   
       }
       if(order[2]!=0) {
	 message(order[1], 0.0, 0.0);
         order[0]=-order[0];
  /* to insert the order in the book in increasing order. 
   * With add( )the arrayList updates the index */
         if(sellOrderStorehouse.size()==0 ||
            order[0]>=extract(sellOrderStorehouse, (sellOrderStorehouse.size()-1), 0)) 
           sellOrderStorehouse.add(order);
	 else
	   sellOrderStorehouse.add(fillStorehouseIncreasingP(order[0]), order); 
         sharesInSellSide+=order[2];
         if(printing==1) {
           System.out.println("Sell order storehouse:");
           for(int i =0; i<sellOrderStorehouse.size(); i++){
             for(int y=0; y<order.length; y++){
               System.out.print(extract(sellOrderStorehouse, i, y) + " ");
             }      
             System.out.println();
           };
           System.out.println();
         }
       }
     }

  // the agent is buying at the maximun price if price>0
     else if(order[0]>0 && order[3]==0){
       /* if there is a compatible order, this one is removed and the order received  
        * it's inserted in the book. It's not necessary shift rows because
        * with remove() the arrayList updates the index */
       while(sellOrderStorehouse.size()>0 && extract(sellOrderStorehouse, 0, 0)<=order[0]
             && order[2]!=0) {
	 if(order[2]==extract(sellOrderStorehouse, 0, 2)){ 
           count+=order[2];
           executedPrice=extract(sellOrderStorehouse, 0, 0);
	   message(order[1], executedPrice, order[2]);
	   message(extract(sellOrderStorehouse, 0, 1), -executedPrice, order[2]);
	   double[] log=new double[2];
	   log[0]=order[2];
	   log[1]=executedPrice;
	   closingContractLog.add(log);
	   currentMeanPrice+=executedPrice*order[2];
           sellOrderStorehouse.remove(0);
           sharesInSellSide-=order[2];
           order[2]=0;
           if(printing==1) {
             System.out.println("Sell order storehouse:");
             for(int i=0; i<sellOrderStorehouse.size(); i++){
               for(int y=0; y<order.length; y++){
                 System.out.print(extract(sellOrderStorehouse, i, y) + " ");
               }      
               System.out.println();
             };
             System.out.println();
           }
	 }
	 else if(order[2]<extract(sellOrderStorehouse, 0, 2)){
           count+=order[2];
           executedPrice=extract(sellOrderStorehouse, 0, 0);
	   message(order[1], executedPrice, order[2]);
	   message(extract(sellOrderStorehouse, 0, 1), -executedPrice, order[2]);
	   double[] log=new double[2];
	   log[0]=order[2];
	   log[1]=executedPrice;
	   closingContractLog.add(log);
	   currentMeanPrice+=executedPrice*order[2];
	   ((double[]) sellOrderStorehouse.get(0))[2]=extract(sellOrderStorehouse, 0, 2)
                                                      -order[2];
           sharesInSellSide-=order[2];
	   order[2]=0;
           if(printing==1) {
             System.out.println("Sell order storehouse:");
             for(int i=0; i<sellOrderStorehouse.size(); i++){
               for(int y=0; y<order.length; y++){
                 System.out.print(extract(sellOrderStorehouse, i, y) + " ");
               }      
               System.out.println();
             };
             System.out.println();
           }
	 }
	 else if(order[2]>extract(sellOrderStorehouse, 0, 2)){
           count+=extract(sellOrderStorehouse, 0, 2);
           executedPrice=extract(sellOrderStorehouse, 0, 0);
	   message(order[1], executedPrice, extract(sellOrderStorehouse, 0, 2));
	   message(extract(sellOrderStorehouse, 0, 1), -executedPrice,
                     extract(sellOrderStorehouse, 0, 2));
	   double[] log=new double[2];
	   log[0]=extract(sellOrderStorehouse, 0, 2);
	   log[1]=executedPrice;
	   closingContractLog.add(log);
	   currentMeanPrice+=executedPrice*extract(sellOrderStorehouse, 0, 2);
	   order[2]=order[2]-extract(sellOrderStorehouse, 0, 2);
           sharesInSellSide-=extract(sellOrderStorehouse, 0, 2);
           sellOrderStorehouse.remove(0);
           if(printing==1) {
             System.out.println("Sell order storehouse:");
             for(int i=0; i<sellOrderStorehouse.size(); i++){
               for(int y=0; y<order.length; y++){
                 System.out.print(extract(sellOrderStorehouse, i, y) + " ");
               }      
               System.out.println();
             };
             System.out.println();
           }
	 }
       }
       if(order[2]!=0) {
	 message(order[1], 0.0, 0.0);
  /* to insert the order in the book in decreasing order. 
   * With add()the arrayList updates the index */
         if(buyOrderStorehouse.size()==0 ||
            order[0]<=extract(buyOrderStorehouse, (buyOrderStorehouse.size()-1) ,0)) 
           buyOrderStorehouse.add(order);
	 else
	   buyOrderStorehouse.add(fillStorehouseDecreasingP(order[0]),order);
         sharesInBuySide+=order[2];     
         if(printing==1) {
           System.out.println("Buy order storhouse:");
           for(int i=0; i<buyOrderStorehouse.size(); i++){
             for(int y=0; y<order.length; y++){
               System.out.print(extract(buyOrderStorehouse, i, y) + " ");
             }      
             System.out.println();
           };
           System.out.println();
         }
       }
     }

    // If the agent is selling at the market price     
     else if(order[0]==0 && order[3]==2) {
       if(buyOrderStorehouse.size()==0)
	  message(order[1], 0, 0);
       else{
         while(buyOrderStorehouse.size()>0 && order[2]!=0) {
	   if(order[2]==extract(buyOrderStorehouse, 0, 2)){ 
             count+=order[2];
             executedPrice=extract(buyOrderStorehouse, 0, 0);
	     message(order[1], -executedPrice, order[2]);
	     message(extract(buyOrderStorehouse, 0, 1), executedPrice, order[2]);
	     double[] log=new double[2];
	     log[0]=order[2];
	     log[1]=executedPrice;
	     closingContractLog.add(log);
	     currentMeanPrice+=executedPrice*order[2];
             buyOrderStorehouse.remove(0);
             sharesInBuySide-=order[2];
	     order[2]=0;
             if(printing==1) {
               System.out.println("Buy order storehouse:");
               for(int i=0; i<buyOrderStorehouse.size(); i++){
                 for(int y=0; y<order.length; y++){
                   System.out.print(extract(buyOrderStorehouse, i, y) + " ");
                 }      
                 System.out.println();
               };
               System.out.println();
             }
	   }
	   else if(order[2]<extract(buyOrderStorehouse, 0, 2)){
             count+=order[2];
             executedPrice=extract(buyOrderStorehouse, 0, 0);
	     message(order[1], -executedPrice, order[2]);
	     message(extract(buyOrderStorehouse, 0, 1), executedPrice, order[2]);
	     double[] log=new double[2];
	     log[0]=order[2];
	     log[1]=executedPrice;
	     closingContractLog.add(log);
	     currentMeanPrice+=executedPrice*order[2];
	     ((double[]) buyOrderStorehouse.get(0))[2]=extract(buyOrderStorehouse, 0, 2)
                                                     -order[2];
             sharesInBuySide-=order[2];
	     order[2]=0;
             if(printing==1) {
               System.out.println("Buy order storehouse:");
               for(int i=0; i<buyOrderStorehouse.size(); i++){
                 for(int y=0; y<order.length; y++){
                   System.out.print(extract(buyOrderStorehouse, i, y) + " ");
                 }      
                 System.out.println();
               };
               System.out.println();
             }
	   }
	   else if(order[2]>extract(buyOrderStorehouse, 0, 2)){
             count+=extract(buyOrderStorehouse, 0, 2);
             executedPrice=extract(buyOrderStorehouse, 0, 0);
	     message(order[1], -executedPrice, extract(buyOrderStorehouse, 0, 2));
	     message(extract(buyOrderStorehouse, 0, 1), executedPrice,
                     extract(buyOrderStorehouse, 0, 2));
	     double[] log=new double[2];
	     log[0]=extract(buyOrderStorehouse, 0, 2);
	     log[1]=executedPrice;
	     closingContractLog.add(log);
	     currentMeanPrice+=executedPrice*extract(buyOrderStorehouse, 0, 2);
	     order[2]=order[2]-extract(buyOrderStorehouse, 0, 2);
             sharesInBuySide-=extract(buyOrderStorehouse, 0, 2);
             buyOrderStorehouse.remove(0);
             if(printing==1) {
               System.out.println("Buy order storehouse:");
               for(int i=0; i<buyOrderStorehouse.size(); i++){
                 for(int y=0; y<order.length; y++){
                   System.out.print(extract(buyOrderStorehouse, i, y) + " ");
                 }      
                 System.out.println();
               };
               System.out.println();
             }
	   }
	 }   
       }
     }

    // the agent is buying at the market price 
     else if(order[0]==0 && order[3]==1){
       if(sellOrderStorehouse.size()==0)
	  message(order[1], 0, 0);
       else{
         while(sellOrderStorehouse.size()>0 && order[2]!=0) {
           if(order[2]==extract(sellOrderStorehouse, 0, 2)){ 
             count+=order[2];
             executedPrice=extract(sellOrderStorehouse, 0, 0);
             message(order[1], executedPrice, order[2]);
             message(extract(sellOrderStorehouse, 0, 1), -executedPrice, order[2]);
             double[] log=new double[2];
             log[0]=order[2];
             log[1]=executedPrice;
             closingContractLog.add(log);
             currentMeanPrice+=executedPrice*order[2];
             sellOrderStorehouse.remove(0);
             sharesInSellSide-=order[2];
             order[2]=0;
             if(printing==1) {
               System.out.println("Sell order storehouse:");
               for(int i=0; i<sellOrderStorehouse.size(); i++){
                 for(int y=0; y<order.length; y++){
                   System.out.print(extract(sellOrderStorehouse, i, y) + " ");
                 }      
                 System.out.println();
               };
               System.out.println();
             }
           }
           else if(order[2]<extract(sellOrderStorehouse, 0, 2)){
             count+=order[2];
             executedPrice=extract(sellOrderStorehouse, 0, 0);
             message(order[1], executedPrice, order[2]);
             message(extract(sellOrderStorehouse, 0, 1), -executedPrice, order[2]);
             double[] log=new double[2];
             log[0]=order[2];
             log[1]=executedPrice;
             closingContractLog.add(log);
             currentMeanPrice+=executedPrice*order[2];
             ((double[]) sellOrderStorehouse.get(0))[2]=extract(sellOrderStorehouse, 0, 2)
                                                      -order[2];
             sharesInSellSide-=order[2];
             order[2]=0;
             if(printing==1) {
               System.out.println("Sell order storehouse:");
               for(int i=0; i<sellOrderStorehouse.size(); i++){
                 for(int y=0; y<order.length; y++){
                   System.out.print(extract(sellOrderStorehouse, i, y) + " ");
                 }      
                 System.out.println();
               };
               System.out.println();
             }
           }
           else if(order[2]>extract(sellOrderStorehouse, 0, 2)){
             count+=extract(sellOrderStorehouse, 0, 2);
             executedPrice=extract(sellOrderStorehouse, 0, 0);
             message(order[1], executedPrice, extract(sellOrderStorehouse, 0, 2));
             message(extract(sellOrderStorehouse, 0, 1), -executedPrice,
                   extract(sellOrderStorehouse, 0, 2));
             double[] log=new double[2];
             log[0]=extract(sellOrderStorehouse, 0, 2);
             log[1]=executedPrice;
             closingContractLog.add(log);
             currentMeanPrice+=executedPrice*extract(sellOrderStorehouse, 0, 2);
             order[2]=order[2]-extract(sellOrderStorehouse, 0, 2);
             sharesInSellSide-=extract(sellOrderStorehouse, 0, 2);
             sellOrderStorehouse.remove(0);
             if(printing==1) {
               System.out.println("Sell order storehouse:");
               for(int i=0; i<sellOrderStorehouse.size(); i++){
                 for(int y=0; y<order.length; y++){
                   System.out.print(extract(sellOrderStorehouse, i, y) + " ");
                 }      
                 System.out.println();
               };
               System.out.println();
	     }
           }
         }
       }
     }
   }

  /** To get the last executed price. */
  public double getPrice(){
    return(executedPrice);
  }

  /** To get the mean price of yesterday. */
  public double getMeanPrice(){
    return(meanPrice);
  }

  /** To get the number of shares filled in the sell side of book. */
  public int getSharesInSellSide(){
    return(sharesInSellSide);
  }

  /** To get the number of shares filled in the buy side of book. */
  public int getSharesInBuySide(){
    return(sharesInBuySide);
  }

  /** To get the closing price of yesterday. */
  public double getPreviousClosingPrice(){
    return(previousClosingPrice);
  }

  /** To get the spread between the first and the last price in the sell side. */
  public double getSellFirstLastSpread(){
    double spread=0;
    if(sellOrderStorehouse.size()==0)
      spread=0;
    else
      spread=extract(sellOrderStorehouse, 0, 0)-
	  extract(sellOrderStorehouse, sellOrderStorehouse.size()-1, 0);
    return(spread);
  }

  /** To get the spread between the first and the last price in the buy side. */
  public double getBuyFirstLastSpread(){
    double spread=0;
    if(buyOrderStorehouse.size()==0)
      spread=0;
    else
      spread=extract(buyOrderStorehouse, 0, 0)-
	  extract(buyOrderStorehouse, buyOrderStorehouse.size()-1, 0);
    return(spread);
  }

  /** To get the spread between first bid and first ask price. */
  public double getSpread(){
    double spread=0;
    if(sellOrderStorehouse.size()==0 && buyOrderStorehouse.size()>0)
      spread=0-extract(buyOrderStorehouse, 0, 0);
    else if(sellOrderStorehouse.size()==0 && buyOrderStorehouse.size()==0)
      spread=0;
    else if(sellOrderStorehouse.size()>0 && buyOrderStorehouse.size()==0)
      spread=extract(sellOrderStorehouse, 0, 0)-0;
    else
    spread=extract(sellOrderStorehouse, 0, 0)-extract(buyOrderStorehouse, 0, 0);
    return(spread);
  }

  /** To get the spread between quantities in sell and buy side. */
  public int getBuySellQuantitySpread(){
    return(sharesInBuySide-sharesInSellSide);
  }

  /** To get the number of shares in sell side in opening or closing Auctions. */
  public int getSharesInSellSideInOpeningOrClosingAuctions(){
    return(sharesInSellSideInOpeningOrClosingAuctions);
  }

  /** To get the number of shares in buy side in opening or closing Auctions. */
  public int getSharesInBuySideInOpeningOrClosingAuctions(){
    return(sharesInBuySideInOpeningOrClosingAuctions);
  }

  /** to get the lagged Mean Price. */
  public double getLaggedMeanPrice(int lag){
    return meanPriceHistory[lag-1];
  }

  /** to get the local History. */
  public int getLocalHistory(){
    return getLocal(localHistory);
  }
}
